/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. See
 * http://www.eu-egee.org/partners/ for details on the copyright holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package org.glite.security.util.namespace;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.ParseException;

import org.glite.security.util.DNHandler;

/**
 * Legacy policy language format.
 * 
 * # Globus CA rights 
 * 
 * access_id_CA X509 '/C=US/O=Globus/CN=Globus Certification Authority'
 * 
 * pos_rights globus CA:sign
 * 
 * cond_subjects globus '"/C=us/O=Globus/*"  "/C=US/O=Globus/*"'
 * 
 * @author alyu
 */
public class LegacyNamespaceFormat extends NamespaceFormat {
//    private static final Logger LOGGER = Logger.getLogger(LegacyNamespaceFormat.class);

    /** Issuer/CA (access identity type) */
    private static final String ACCESS_ID_CA = "access_id_ca";

    /** Subject DN positive rights */
    private static final String POS_RIGHTS = "pos_rights";

    /** Subject (access identity type) */
    private static final String COND_SUBJECTS = "cond_subjects";

    /**
     * Temp place holder used during parsing.
     */
    private String issuer = null;

    /**
     * Temp place holder used during parsing.
     */
    private String accessRights = null;

    /**
     * The constructor to create an empty namespace.
     */
    public LegacyNamespaceFormat() {
    	//empty on purpose
    }

    /**
     * Parses a namespaces file.
     * 
     * @param filename filename of the namespaces file
     * @throws IOException if unsuccessful
     * @throws ParseException if the namespaces file format is incorrect
     */
    public void parse(String filename) throws IOException, ParseException {
        File file = new File(filename);
        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new FileReader(file));

            String line = null;
            int lineNumber = 0;
            while ((line = reader.readLine()) != null) {
                line = line.trim(); // remove any white spaces
                lineNumber++;
                // ignore comments
                if (!line.contains("#") && line.length() > 0) {
                    // Create a new policy pair
                    addPolicy(line, lineNumber, filename);
                }
            } // while (
        } catch (IOException e) {
            throw e;
        } catch (ParseException e) {
            throw e;
        } finally {
            if (reader != null) {
                reader.close();
            }
        }
    }

    /**
     * Creates a new policy pair.
     *
     * @param inputStatement policy statement to parse.
     * @param lineNumber the line number of the policy statement, for error messages.
     * @param filename The filename of the policy definitions, for error messages.
     * @throws ParseException if policy parsing is unsuccessful.
     */
    @SuppressWarnings("unused")
	private void addPolicy(String inputStatement, int lineNumber, String filename) throws ParseException {
    	String statement = inputStatement.toLowerCase(); // make everything to lower case

        if (statement.contains(ACCESS_ID_CA)) {
            // LOGGER.debug("Line-->" + line);
            statement = statement.trim();
            String[] split = statement.split(ACCESS_ID_CA + "\\s+x509\\s+");

            issuer = split[1].replace("'", "").replace("\"", "").trim();

        } else if (statement.contains(POS_RIGHTS)) {
            String[] split = statement.split(POS_RIGHTS + "\\s+globus\\s+");
            accessRights = split[1].replace("'", "").replace("\"", "").trim();
        } else if (statement.contains(COND_SUBJECTS)) {
            String[] split = statement.split(COND_SUBJECTS + "\\s+globus");

            if (issuer != null) {
                String[] subjects = split[1].split("['\".\"]|[\".\"]|[\".\"']");

                for (String subject : subjects) {
                    if (subject.trim().length() > 0) {
                        NamespacePolicy policy = new NamespacePolicy(DNHandler.getDNRFC2253(issuer), accessRights, subject.trim(), statement,
                                lineNumber, filename);
                        getPolices().add(policy);
                    }
                }
            }
            issuer = null;
            accessRights = null;
        } else {
            // ignore
        }
    }
}
